/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: OSGiResourceInjector.java 2144 2007-12-10 19:43:59Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.osgi.extension;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.osgi.framework.BundleContext;
import org.ow2.easybeans.api.bean.EasyBeansBean;
import org.ow2.easybeans.api.injection.ResourceInjector;
import org.ow2.easybeans.osgi.annotation.OSGiResource;

/**
 * Injects some OSGi resources in the given bean.
 * @author Florent BENOIT
 */
public class OSGiResourceInjector implements ResourceInjector {

    /**
     * Instance of a bundle context.
     */
    private BundleContext bundleContext;

    /**
     * Default constructor.
     */
    public OSGiResourceInjector() {

    }

    /**
     * Build an injector with a given bundlecontext.
     * @param bundleContext the given context
     */
    public OSGiResourceInjector(final BundleContext bundleContext) {
        setBundleContext(bundleContext);
    }

    /**
     * Gets the bundle context.
     * @return the bundle context
     */
    public BundleContext getBundleContext() {
        return bundleContext;
    }

    /**
     * Sets the given bundle context.
     * @param bundleContext the given context
     */
    public void setBundleContext(final BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    /**
     * Called <b>before</b> EasyBeans resolution and injection of dependencies.
     * @param bean the Bean instance.
     */
    public void preEasyBeansInject(final EasyBeansBean bean) {

        // Process field
        processFields(bean);

        // Then methods
        processMethods(bean);

    }

    /**
     * Do injection on the bean fields.
     * @param bean the given bean
     */
    protected void processFields(final EasyBeansBean bean) {
        // Search all fields of this instance
        Field[] fields = bean.getClass().getDeclaredFields();
        if (fields == null) {
            return;
        }

        // For each field
        for (Field field : fields) {
            if (field.isAnnotationPresent(OSGiResource.class)) {
                Class<?> fieldClassType = field.getType();
                if (BundleContext.class.getName().equals(fieldClassType.getName())) {
                    injectField(field, bean, bundleContext);
                }
            }
        }
    }

    /**
     * Do injection on the selected methods.
     * @param bean the given bean
     */
    protected void processMethods(final EasyBeansBean bean) {
        // Search all methods of this instance
        Method[] methods = bean.getClass().getDeclaredMethods();
        if (methods == null) {
            return;
        }

        // For each method
        for (Method method : methods) {
            if (method.isAnnotationPresent(OSGiResource.class)) {

                // Setter with one parameter
                if (!method.getName().startsWith("set") || method.getParameterTypes().length != 1) {
                    continue;
                }

                // Get Class of the method
                Class<?> parameterClass = method.getParameterTypes()[0];
                if (BundleContext.class.getName().equals(parameterClass.getName())) {
                    invokeMethod(method, bean, bundleContext);
                }
            }
        }
    }

    /**
     * Inject the given value in the instance for the selected field.
     * @param <T> type of the value
     * @param field the field to use
     * @param bean the instance that will be used for injection
     * @param value the given value to set
     */
    protected <T> void injectField(final Field field, final EasyBeansBean bean, final T value) {
        boolean isAccessible = field.isAccessible();
        try {
            field.setAccessible(true);
            try {
                field.set(bean, value);
            } catch (IllegalArgumentException e) {
                throw new IllegalStateException("Unable to inject value '" + value + "' in the bean '" + bean
                        + "' for the field '" + field + "'.", e);
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Unable to inject value '" + value + "' in the bean '" + bean
                        + "' for the field '" + field + "'.", e);
            }
        } finally {
            field.setAccessible(isAccessible);
        }

    }

    /**
     * Sets the value on the given method of the given instance.
     * @param <T> the type of the value
     * @param method the method to call
     * @param bean the instance on which the method will be called
     * @param value the value to set
     */
    protected <T> void invokeMethod(final Method method, final EasyBeansBean bean, final T value) {
        boolean isAccessible = method.isAccessible();
        try {
            method.setAccessible(true);
            try {
                method.invoke(bean, value);
            } catch (IllegalArgumentException e) {
                throw new IllegalStateException("Unable to call setter method '" + method + "' in the bean '" + bean
                        + "' for the value '" + value + "'.", e);
            } catch (IllegalAccessException e) {
                throw new IllegalStateException("Unable to call setter method '" + method + "' in the bean '" + bean
                        + "' for the value '" + value + "'.", e);
            } catch (InvocationTargetException e) {
                throw new IllegalStateException("Unable to call setter method '" + method + "' in the bean '" + bean
                        + "' for the value '" + value + "'.", e);
            }
        } finally {
            method.setAccessible(isAccessible);
        }

    }

    /**
     * Called <b>after</b> EasyBeans resolution and injection of dependencies.
     * @param bean the Bean instance.
     */
    public void postEasyBeansInject(final EasyBeansBean bean) {
        // Do nothing
    }
}
